From c47f8c0d6ca82f6b4525066bbd53f42775f8d1d9 Mon Sep 17 00:00:00 2001
From: Andrew Price <andrewprice@andrewalexanderprice.com>
Date: Sat, 16 Jul 2011 02:07:56 +0930
Subject: [PATCH 3/8] Fixed bugs with 3d vision texture not stretching across entire window.

---
 STEREOSCOPY README                  |    2 +
 configure.ac                        |    2 +-
 include/vlc_es.h                    |   28 ++
 include/vlc_picture.h               |   11 +
 include/vlc_vout_display.h          |    4 +
 modules/gui/qt4/menus.cpp           |    2 +
 modules/video_filter/Modules.am     |    2 +
 modules/video_filter/stereoscopy.c  |    4 +-
 modules/video_output/msw/common.h   |   49 ++++
 modules/video_output/msw/direct3d.c |  537 +++++++++++++++++++++++++++++++----
 10 files changed, 586 insertions(+), 55 deletions(-)

diff --git a/STEREOSCOPY README b/STEREOSCOPY README
index c91c0a5..249cf16 100644
--- a/STEREOSCOPY README	
+++ b/STEREOSCOPY README	
@@ -18,6 +18,8 @@ cp -v /usr/win32/share/aclocal/* m4/
 cp -v /usr/share/aclocal/* m4/
 PATH=/usr/win32/bin:$PATH ./bootstrap
 
+if it fails witH
+
 ================================================================================
 2. Then to configure:
 
diff --git a/configure.ac b/configure.ac
index 036f502..b3ae94d 100644
--- a/configure.ac
+++ b/configure.ac
@@ -4255,7 +4255,7 @@ AC_SUBST(VERSION_REVISION)
 AC_SUBST(VERSION_EXTRA)
 AC_SUBST(COPYRIGHT_YEARS)
 AC_DEFINE_UNQUOTED(VLC_COMPILE_BY, "`whoami`", [user who ran configure])
-AC_DEFINE_UNQUOTED(VLC_COMPILE_HOST, "`hostname -f 2>/dev/null || hostname`", [host which ran configure])
+AC_DEFINE_UNQUOTED(VLC_COMPILE_HOST, "`hostname`", [host which ran configure])
 AC_DEFINE_UNQUOTED(VLC_COMPILER, "`$CC -v 2>&1 | tail -n 1`", [compiler])
 dnl Win32 need s a numerical version_extra.
 case $( echo ${VERSION_EXTRA}|wc -m ) in
diff --git a/include/vlc_es.h b/include/vlc_es.h
index 5bf619e..4f50e7e 100644
--- a/include/vlc_es.h
+++ b/include/vlc_es.h
@@ -98,6 +98,34 @@ struct audio_format_t
 /**
  * video format description
  */
+ 
+/**
+ * stereoscopic formats
+ * sbs = side by side
+ * anaglyph = each eye is encoded in a colour channel
+ * field sequential = each eye is stored in alternating frames
+ */
+#define VIDEO_FORMAT_STEREOSCOPY_2D (0)                            /** no 3d */
+#define VIDEO_FORMAT_STEREOSCOPY_SBS_FULL_HORIZONTAL (1)
+#define VIDEO_FORMAT_STEREOSCOPY_SBS_HALF_HORIZONTAL (2)
+#define VIDEO_FORMAT_STEREOSCOPY_SBS_FULL_VERTICAL (3)
+#define VIDEO_FORMAT_STEREOSCOPY_SBS_HALF_VERTICAL (4)
+#define VIDEO_FORMAT_STEREOSCOPY_ANAGLYPH_RED_GREEN (5)
+#define VIDEO_FORMAT_STEREOSCOPY_ANAGLYPH_RED_BLUE (6)
+#define VIDEO_FORMAT_STEREOSCOPY_ANAGLYPH_RED_CYAN (7)
+#define VIDEO_FORMAT_STEREOSCOPY_ANAGLYPH_GREEN_BLUE (8)
+#define VIDEO_FORMAT_STEREOSCOPY_ANAGLYPH_GREEN_MAGENTA (9)
+#define VIDEO_FORMAT_STEREOSCOPY_ANAGLYPH_BLUE_YELLOW (10)
+#define VIDEO_FORMAT_STEREOSCOPY_FIELD_SEQUENTIAL (11)
+                     /** fill in the missing colour channels with the average
+					                                   of the other channels */
+											     
+#define VIDEO_FORMAT_STEREOSCOPY_ANAGLYPH_FILL_CHANNELS (0x20000000)
+                                             /** swap the colour of each eye */
+#define VIDEO_FORMAT_STEREOSCOPY_ANAGLYPH_SWAP_EYES (0x40000000)
+#define VIDEO_FORMAT_STEREOSCOPY_SWAP_EYES (0x80000000)        /** flip eyes */
+/*    unsigned int i_stereoscopy;              *//**stereoscopy format*/
+
 struct video_format_t
 {
     vlc_fourcc_t i_chroma;                               /**< picture chroma */
diff --git a/include/vlc_picture.h b/include/vlc_picture.h
index f264d76..5387e38 100644
--- a/include/vlc_picture.h
+++ b/include/vlc_picture.h
@@ -75,6 +75,15 @@ struct picture_t
     plane_t         p[PICTURE_PLANE_MAX];     /**< description of the planes */
     int             i_planes;                /**< number of allocated planes */
 
+    /** \name Stereoscopy properties
+     **@{*/
+    int             i_eye;                /**< the stereoscopy eye, may be:
+                                                0 - 2d image,
+                                                1- left eye,
+                                                2 - right eye - could possibly be > 2 for more
+                                                    than 2 eyes */
+    /**@}*/
+
     /** \name Picture management properties
      * These properties can be modified using the video output thread API,
      * but should never be written directly */
@@ -100,6 +109,7 @@ struct picture_t
      * keep track of the picture */
     picture_sys_t * p_sys;
 
+
     /** This way the picture_Release can be overloaded */
     void (*pf_release)( picture_t * );
     picture_release_sys_t *p_release_sys;
@@ -216,6 +226,7 @@ static inline void picture_CopyProperties( picture_t *p_dst, const picture_t *p_
     p_dst->i_nb_fields = p_src->i_nb_fields;
     p_dst->b_top_field_first = p_src->b_top_field_first;
 
+    p_dst->i_eye = p_src->i_eye; /* for stereoscopy */
     /* FIXME: copy ->p_q and ->p_qstride */
 }
 
diff --git a/include/vlc_vout_display.h b/include/vlc_vout_display.h
index 77d66e2..c5bae9d 100644
--- a/include/vlc_vout_display.h
+++ b/include/vlc_vout_display.h
@@ -171,6 +171,10 @@ enum {
      * video_format_t::i_visible_width/height */
     VOUT_DISPLAY_CHANGE_SOURCE_CROP,   /* const video_format_t *p_source */
 
+	/* Ask the module to acknowledge/refuse source stereoscopy format change after being
+	 * requested externally */
+    VOUT_DISPLAY_CHANGE_STEREOSCOPY,
+
     /* Ask an opengl interface if available. */
     VOUT_DISPLAY_GET_OPENGL,           /* vlc_gl_t ** */
 };
diff --git a/modules/gui/qt4/menus.cpp b/modules/gui/qt4/menus.cpp
index be893ae..a8653b6 100644
--- a/modules/gui/qt4/menus.cpp
+++ b/modules/gui/qt4/menus.cpp
@@ -239,6 +239,7 @@ static int VideoAutoMenuBuilder( vout_thread_t *p_object,
     PUSH_VAR( "deinterlace" );
     PUSH_VAR( "deinterlace-mode" );
     PUSH_VAR( "postprocess" );
+    PUSH_VAR( "stereoscopy" );
 
     return VLC_SUCCESS;
 }
@@ -651,6 +652,7 @@ QMenu *QVLCMenu::VideoMenu( intf_thread_t *p_intf, QMenu *current, bool b_subtit
         addActionWithSubmenu( current, "deinterlace", qtr( "&Deinterlace" ) );
         addActionWithSubmenu( current, "deinterlace-mode", qtr( "&Deinterlace mode" ) );
         addActionWithSubmenu( current, "postprocess", qtr( "&Post processing" ) );
+        addActionWithSubmenu( current, "stereoscopy", qtr( "&Stereoscopy" ) );
     }
 
     p_input = THEMIM->getInput();
diff --git a/modules/video_filter/Modules.am b/modules/video_filter/Modules.am
index e6abb4f..a86fb20 100644
--- a/modules/video_filter/Modules.am
+++ b/modules/video_filter/Modules.am
@@ -57,6 +57,7 @@ SOURCES_postproc = postproc.c
 SOURCES_swscale = swscale.c ../codec/avcodec/chroma.c
 SOURCES_scene = scene.c
 SOURCES_sepia = sepia.c
+SOURCES_stereoscopy = stereoscopy.c
 SOURCES_yuvp = yuvp.c
 SOURCES_antiflicker = antiflicker.c
 SOURCES_atmo = atmo/atmo.cpp \
@@ -124,6 +125,7 @@ libvlc_LTLIBRARIES += \
 	libsepia_plugin.la \
 	libsharpen_plugin.la \
 	libsubsdelay_plugin.la \
+	libstereoscopy_plugin.la \
 	libtransform_plugin.la \
 	libwall_plugin.la \
 	libwave_plugin.la \
diff --git a/modules/video_filter/stereoscopy.c b/modules/video_filter/stereoscopy.c
index 5e5114f..cd96e81 100644
--- a/modules/video_filter/stereoscopy.c
+++ b/modules/video_filter/stereoscopy.c
@@ -499,7 +499,7 @@ static picture_t *Filter( filter_t *p_filter, picture_t *p_inpic )
             return NULL;
         }
 
-        p_rightOut->i_eye = 1;
+        p_rightOut->i_eye = 2;
         picture_Release( p_inpic );
         return p_rightOut;
     }
@@ -518,7 +518,7 @@ static picture_t *Filter( filter_t *p_filter, picture_t *p_inpic )
             return NULL;
         }
 
-        p_leftOut->i_eye = 2;
+        p_leftOut->i_eye = 1;
         picture_Release( p_inpic );
         return p_leftOut;
     }
diff --git a/modules/video_output/msw/common.h b/modules/video_output/msw/common.h
index 0a8fbae..8a71a64 100644
--- a/modules/video_output/msw/common.h
+++ b/modules/video_output/msw/common.h
@@ -28,6 +28,11 @@
 #include <vlc_picture_pool.h>
 #include "events.h"
 
+#ifdef MODULE_NAME_IS_direct3d
+/* assume d3d9.h has alreayd been included */
+#include <nvapi.h>
+#endif
+
 #ifdef MODULE_NAME_IS_wingapi
     typedef struct GXDisplayProperties {
         DWORD cxWidth;
@@ -62,6 +67,23 @@
 
 #endif
 
+
+#ifdef MODULE_NAME_IS_direct3d
+    /* The off-screen header that defines a stereo texture. */
+#   define NVSTEREO_IMAGE_SIGNATURE 0x4433564e /* "NV3D" */
+    typedef struct DXNvStereoImageHeader
+    {
+        DWORD dwSignature;
+        DWORD dwWidth;
+        DWORD dwHeight;
+        DWORD dwBPP;
+        DWORD dwFlags;
+    } NVSTEREOIMAGEHEADER, *LPNVSTEREOIMAGEHEADER;
+/* ORed flags in the dwFlags fiels of the NVSTEREO_IMAGE_SIGNATURE structure above */
+#   define DXNV_SWAP_EYES    0x01
+#   define DXNV_SCALE_TO_FIT 0x02
+#endif
+
 /*****************************************************************************
  * vout_sys_t: video output method descriptor
  *****************************************************************************
@@ -179,11 +201,38 @@ struct vout_display_sys_t
     D3DCAPS9                d3dcaps;
     LPDIRECT3DDEVICE9       d3ddev;
     D3DPRESENT_PARAMETERS   d3dpp;
+    
+    /* nvapi objects */
+    HINSTANCE               hNvAPIWrapper_dll;
+    bool                    use_nvidia_3d; /* use nVidia 3D Vision - is it
+                                              loaded? if false then can
+                                              assume dll is not loaded */
+    void                    *nvStereoHandle; /* 3d vision stereo handle */
+    void                    *nvInitializeFunc; /* NvAPI_Initialize */
+    void                    *nvUnloadFunc; /* NvAPI_Unload */
+    void                    *nvStereoIsEnabledFunc; /* NvAPI_Stereo_IsEnabled */
+    void                    *nvStereoEnableFunc ; /* NvAPI_Stereo_Enable */
+    void                    *nvStereoCreateHandleFunc; /* NvAPI_Stereo_CreateHandleFromIUnknown */
+    void                    *nvStereoIsActivatedFunc; /* NvAPI_Stereo_IsActivated */
+    void                    *nvStereoDestroyHandleFunc; /* NvAPI_Stereo_DestroyHandle */
+
+    RECT                    nvRectLeft;
+    RECT                    nvRectRight;
+    RECT                    nvRectHeader;
+
     // scene objects
     LPDIRECT3DTEXTURE9      d3dtex;
     LPDIRECT3DVERTEXBUFFER9 d3dvtc;
     int                     d3dregion_count;
     struct d3d_region_t     *d3dregion;
+	/* Stereoscopic scene objects */
+    LPDIRECT3DTEXTURE9      d3dtex_right; /* store right eye, d3dtex for left */
+    LPDIRECT3DSURFACE9      d3dnv_surface;
+    LPDIRECT3DSURFACE9      d3dback_buffer;
+    LPDIRECT3DSURFACE9      d3dleft_surface;
+	LPDIRECT3DSURFACE9      d3dright_surface;
+	bool                    left_tex_filled;
+	bool                    right_tex_filled;
 
     picture_resource_t      resource;
 
diff --git a/modules/video_output/msw/direct3d.c b/modules/video_output/msw/direct3d.c
index da11e58..bf2d0cf 100644
--- a/modules/video_output/msw/direct3d.c
+++ b/modules/video_output/msw/direct3d.c
@@ -73,6 +73,7 @@ vlc_module_begin ()
 
     add_bool("direct3d-desktop", false, DESKTOP_TEXT, DESKTOP_LONGTEXT, true)
     add_bool("direct3d-hw-blending", true, HW_BLENDING_TEXT, HW_BLENDING_LONGTEXT, true)
+    
 
     set_capability("vout display", 240)
     add_shortcut("direct3d")
@@ -115,6 +116,9 @@ static void Direct3DDestroy(vout_display_t *);
 static int  Direct3DOpen (vout_display_t *, video_format_t *);
 static void Direct3DClose(vout_display_t *);
 
+static int NvCreate  (vout_display_t *);
+static void NvDestroy (vout_display_t *);
+
 /* */
 typedef struct
 {
@@ -135,7 +139,7 @@ typedef struct d3d_region_t {
 
 static void Direct3DDeleteRegions(int, d3d_region_t *);
 
-static int  Direct3DImportPicture(vout_display_t *vd, d3d_region_t *, LPDIRECT3DSURFACE9 surface);
+static int  Direct3DImportPicture(vout_display_t *vd, d3d_region_t *, LPDIRECT3DSURFACE9 surface, int i_eye);
 static void Direct3DImportSubpicture(vout_display_t *vd, int *, d3d_region_t **, subpicture_t *);
 
 static void Direct3DRenderScene(vout_display_t *vd, d3d_region_t *, int, d3d_region_t *);
@@ -148,6 +152,10 @@ static int DesktopCallback(vlc_object_t *, char const *, vlc_value_t, vlc_value_
  */
 static int Open(vlc_object_t *object)
 {
+    /* Temporary warning */
+    MessageBoxA(NULL, "This build of VLC is uses priority code from NVAPI and is not to be redistributed to binary format.",
+        "Licensing Warning", MB_OK);
+
     vout_display_t *vd = (vout_display_t *)object;
     vout_display_sys_t *sys;
 
@@ -156,9 +164,12 @@ static int Open(vlc_object_t *object)
     if (!sys)
         return VLC_ENOMEM;
 
+    NvCreate(vd);
+
     if (Direct3DCreate(vd)) {
         msg_Err(vd, "Direct3D could not be initialized");
         Direct3DDestroy(vd);
+        NvDestroy(vd);
         free(sys);
         return VLC_EGENERIC;
     }
@@ -230,6 +241,7 @@ error:
     Direct3DClose(vd);
     CommonClean(vd);
     Direct3DDestroy(vd);
+    NvDestroy(vd);
     free(vd->sys);
     return VLC_EGENERIC;
 }
@@ -268,6 +280,7 @@ static void Close(vlc_object_t *object)
     CommonClean(vd);
 
     Direct3DDestroy(vd);
+    NvDestroy(vd);
 
     free(vd->sys);
 }
@@ -286,6 +299,7 @@ static void Prepare(vout_display_t *vd, picture_t *picture, subpicture_t *subpic
 {
     vout_display_sys_t *sys = vd->sys;
     LPDIRECT3DSURFACE9 surface = picture->p_sys->surface;
+    int i_eye = picture->i_eye; /* which eye is this picture from */
 #if 0
     picture_Release(picture);
     VLC_UNUSED(subpicture);
@@ -311,7 +325,7 @@ static void Prepare(vout_display_t *vd, picture_t *picture, subpicture_t *subpic
     }
 
     d3d_region_t picture_region;
-    if (!Direct3DImportPicture(vd, &picture_region, surface)) {
+    if (!Direct3DImportPicture(vd, &picture_region, surface, i_eye)) {
         int subpicture_region_count     = 0;
         d3d_region_t *subpicture_region = NULL;
         if (subpicture)
@@ -497,6 +511,130 @@ static void Manage (vout_display_t *vd)
 }
 
 /**
+ * Initialises an instance of 3D vision
+ */
+static int NvCreate(vout_display_t *vd)
+{
+    vout_display_sys_t *sys = vd->sys;
+    sys->hNvAPIWrapper_dll = LoadLibrary(TEXT("nvapiwrapper.dll"));
+    if(!sys->hNvAPIWrapper_dll) {
+        MessageBoxA(NULL, "Cannot find nvapiwrapper.dll. Is it in the VLC directory?",
+        "nvapiwrapper.dll", MB_ICONERROR);
+        sys->use_nvidia_3d = false;
+        return VLC_EGENERIC;
+    }
+
+    /* load in function pointers */
+    sys->nvInitializeFunc = (void *)GetProcAddress(sys->hNvAPIWrapper_dll,
+        TEXT("_NvAPIWrapper_Initialize@0"));
+    if(!sys->nvInitializeFunc) {
+        msg_Err(vd, "Cannot locate NvAPIWrapper_Initialize in DLL");
+        FreeLibrary(sys->hNvAPIWrapper_dll);
+        sys->use_nvidia_3d = false;
+        return VLC_EGENERIC;
+    }
+
+    sys->nvUnloadFunc = (void *)GetProcAddress(sys->hNvAPIWrapper_dll,
+        TEXT("_NvAPIWrapper_Unload@0"));
+    if(!sys->nvUnloadFunc) {
+        msg_Err(vd, "Cannot locate NvAPIWrapper_Unload in DLL");
+        FreeLibrary(sys->hNvAPIWrapper_dll);
+        sys->use_nvidia_3d = false;
+        return VLC_EGENERIC;
+    }
+
+    sys->nvStereoIsEnabledFunc = (void *)GetProcAddress(sys->hNvAPIWrapper_dll,
+        TEXT("_NvAPIWrapper_Stereo_IsEnabled@4"));
+    if(!sys->nvStereoIsEnabledFunc) {
+        msg_Err(vd, "Cannot locate NvAPIWrapper_Stereo_IsEnabled in DLL");
+        FreeLibrary(sys->hNvAPIWrapper_dll);
+        sys->use_nvidia_3d = false;
+        return VLC_EGENERIC;
+    }
+
+    sys->nvStereoEnableFunc = (void *)GetProcAddress(sys->hNvAPIWrapper_dll,
+        TEXT("_NvAPIWrapper_Stereo_Enable@0"));
+    if(!sys->nvStereoEnableFunc) {
+        msg_Err(vd, "Cannot locate NvAPIWrapper_Stereo_Enable in DLL");
+        FreeLibrary(sys->hNvAPIWrapper_dll);
+        sys->use_nvidia_3d = false;
+        return VLC_EGENERIC;
+    }
+
+    sys->nvStereoCreateHandleFunc = (void *)GetProcAddress(sys->hNvAPIWrapper_dll,
+        TEXT("_NvAPIWrapper_Stereo_CreateHandleFromIUnknown@8"));
+    if(!sys->nvStereoCreateHandleFunc) {
+        msg_Err(vd, "Cannot locate NvAPIWrapper_Stereo_CreateHandleFromIUnknown in DLL");
+        FreeLibrary(sys->hNvAPIWrapper_dll);
+        sys->use_nvidia_3d = false;
+        return VLC_EGENERIC;
+    }
+
+    sys->nvStereoIsActivatedFunc = (void *)GetProcAddress(sys->hNvAPIWrapper_dll,
+        TEXT("_NvAPIWrapper_Stereo_IsActivated@8"));
+    if(!sys->nvStereoIsActivatedFunc) {
+        msg_Err(vd, "Cannot locate NvAPIWrapper_Stereo_IsActivated in DLL");
+        FreeLibrary(sys->hNvAPIWrapper_dll);
+        sys->use_nvidia_3d = false;
+        return VLC_EGENERIC;
+    }
+
+    sys->nvStereoDestroyHandleFunc = (void *)GetProcAddress(sys->hNvAPIWrapper_dll,
+        TEXT("_NvAPIWrapper_Stereo_DestroyHandle@4"));
+    if(!sys->nvStereoDestroyHandleFunc) {
+        msg_Err(vd, "Cannot locate NvAPIWrapper_Stereo_DestroyHandle in DLL");
+        FreeLibrary(sys->hNvAPIWrapper_dll);
+        sys->use_nvidia_3d = false;
+        return VLC_EGENERIC;
+    }
+
+    /* now try initialising nvapi */
+    NvAPI_Status (WINAPI *NvAPIWrapper_Initialise)();
+    NvAPIWrapper_Initialise = sys->nvInitializeFunc;
+
+    NvAPI_Status (WINAPI *NvAPIWrapper_Unload)();
+    NvAPIWrapper_Unload = sys->nvUnloadFunc;
+
+    NvAPI_Status (WINAPI *NvAPIWrapper_Stereo_IsEnabled)(NvU8 *pIsStereoEnabled);
+    NvAPIWrapper_Stereo_IsEnabled = sys->nvStereoIsEnabledFunc;
+
+    NvAPI_Status (WINAPI *NvAPIWrapper_Stereo_Enable)(void);
+    NvAPIWrapper_Stereo_Enable = sys->nvStereoEnableFunc;
+
+    if(NvAPIWrapper_Initialise() == NVAPI_OK)
+    {
+        /* not returning NVAPI_OK for Initialise() could mean bad drivers or
+           not using an nVidia card */
+        NvU8 stereoEnabled;
+        if(NvAPIWrapper_Stereo_IsEnabled(&stereoEnabled) != NVAPI_OK)
+        {
+            /* not returning NVAPI_OK for Stereo_IsEnabled could possibly mean
+               an unsupported video card or not running Windows Vista/7 */
+            msg_Err(vd, "NV stereo did not want to enable");
+            NvAPIWrapper_Unload();
+            FreeLibrary(sys->hNvAPIWrapper_dll);
+            sys->use_nvidia_3d = false;
+            return VLC_EGENERIC;
+        }
+        else
+        {
+            if(stereoEnabled == 0)
+                NvAPIWrapper_Stereo_Enable();
+        }
+    }
+    else
+    {
+        msg_Err(vd, "NvAPI did not want to initialize");
+        FreeLibrary(sys->hNvAPIWrapper_dll);
+        sys->use_nvidia_3d = false;
+        return VLC_EGENERIC;
+    }
+
+    sys->use_nvidia_3d = true;
+    return VLC_SUCCESS;
+}
+
+/**
  * It initializes an instance of Direct3D9
  */
 static int Direct3DCreate(vout_display_t *vd)
@@ -541,6 +679,26 @@ static int Direct3DCreate(vout_display_t *vd)
 }
 
 /**
+ * It releases an instance of nVidia 3D Vision
+ */
+static void NvDestroy(vout_display_t *vd)
+{
+    vout_display_sys_t *sys = vd->sys;
+
+    if(!sys->use_nvidia_3d)
+        return;
+
+    sys->use_nvidia_3d = false;
+
+    NvAPI_Status (WINAPI *NvAPIWrapper_Unload)();
+    NvAPIWrapper_Unload = sys->nvUnloadFunc;
+
+    NvAPIWrapper_Unload();
+
+    FreeLibrary(sys->hNvAPIWrapper_dll);
+}
+
+/**
  * It releases an instance of Direct3D9
  */
 static void Direct3DDestroy(vout_display_t *vd)
@@ -653,15 +811,32 @@ static int Direct3DOpen(vout_display_t *vd, video_format_t *fmt)
 
     UpdateRects(vd, NULL, NULL, true);
 
+    /* attempt to create a d3d device */
+    if(sys->use_nvidia_3d)
+    {
+        NvAPI_Status (WINAPI *NvAPIWrapper_Stereo_CreateHandleFromIUnknown)
+            (IUnknown *pDevice, StereoHandle *pStereoHandle);
+        NvAPIWrapper_Stereo_CreateHandleFromIUnknown =
+            sys->nvStereoCreateHandleFunc;
+
+        if(NvAPIWrapper_Stereo_CreateHandleFromIUnknown(sys->d3ddev,
+            &sys->nvStereoHandle) != NVAPI_OK)
+        {
+	    msg_Err(vd, "Could not create a stereo handle.");
+            NvDestroy(vd);
+        }
+    }
+
     if (Direct3DCreateResources(vd, fmt)) {
         msg_Err(vd, "Failed to allocate resources");
         return VLC_EGENERIC;
     }
-
+    
     /* Change the window title bar text */
     EventThreadUpdateTitle(sys->event, VOUT_TITLE " (Direct3D output)");
 
     msg_Dbg(vd, "Direct3D device adapter successfully initialized");
+    
     return VLC_SUCCESS;
 }
 
@@ -674,6 +849,15 @@ static void Direct3DClose(vout_display_t *vd)
 
     Direct3DDestroyResources(vd);
 
+    if(sys->use_nvidia_3d)
+    {
+        /* destroy stereo handle */
+        NvAPI_Status (WINAPI *NvAPIWrapper_Stereo_DestroyHandle)(
+            StereoHandle stereoHandle);
+        NvAPIWrapper_Stereo_DestroyHandle = sys->nvStereoDestroyHandleFunc;
+        NvAPIWrapper_Stereo_DestroyHandle(sys->nvStereoHandle);
+    }
+
     if (sys->d3ddev)
        IDirect3DDevice9_Release(sys->d3ddev);
 
@@ -694,6 +878,15 @@ static int Direct3DReset(vout_display_t *vd)
     /* release all D3D objects */
     Direct3DDestroyResources(vd);
 
+    if(sys->use_nvidia_3d)
+    {
+        /* destroy stereo handle */
+        NvAPI_Status (WINAPI *NvAPIWrapper_Stereo_DestroyHandle)(
+            StereoHandle stereoHandle);
+        NvAPIWrapper_Stereo_DestroyHandle = sys->nvStereoDestroyHandleFunc;
+        NvAPIWrapper_Stereo_DestroyHandle(sys->nvStereoHandle);
+    }
+
     /* */
     HRESULT hr = IDirect3DDevice9_Reset(d3ddev, &sys->d3dpp);
     if (FAILED(hr)) {
@@ -701,6 +894,22 @@ static int Direct3DReset(vout_display_t *vd)
         return VLC_EGENERIC;
     }
 
+    /* attempt to create a d3d device */
+    if(sys->use_nvidia_3d)
+    {
+        NvAPI_Status (WINAPI *NvAPIWrapper_Stereo_CreateHandleFromIUnknown)
+            (IUnknown *pDevice, StereoHandle *pStereoHandle);
+        NvAPIWrapper_Stereo_CreateHandleFromIUnknown =
+            sys->nvStereoCreateHandleFunc;
+
+        if(NvAPIWrapper_Stereo_CreateHandleFromIUnknown(sys->d3ddev,
+            &sys->nvStereoHandle) != NVAPI_OK)
+        {
+	    msg_Err(vd, "Could not create a stereo handle.");
+            NvDestroy(vd);
+        }
+    }
+
     UpdateRects(vd, NULL, NULL, true);
 
     /* re-create them */
@@ -1013,12 +1222,116 @@ static int Direct3DCreateScene(vout_display_t *vd, const video_format_t *fmt)
         return VLC_EGENERIC;
     }
 
+    LPDIRECT3DSURFACE9 d3dleft_surface;
+    hr = IDirect3DTexture9_GetSurfaceLevel(d3dtex, 0, &d3dleft_surface);
+    if (FAILED(hr)) {
+        msg_Err(vd, "Failed to get surface of left texture. (hr=0x%lx)", hr);
+        IDirect3DTexture9_Release(d3dtex);
+        IDirect3DVertexBuffer9_Release(d3dvtc);
+        return VLC_EGENERIC;
+    }
+
+    if (sys->use_nvidia_3d)
+    {
+        /* secondary texture for right eye */
+        LPDIRECT3DTEXTURE9 d3dtex_right;
+        hr = IDirect3DDevice9_CreateTexture(d3ddev,
+                                            fmt->i_width,
+                                            fmt->i_height,
+                                            1,
+                                            D3DUSAGE_RENDERTARGET,
+                                            sys->d3dpp.BackBufferFormat,
+                                            D3DPOOL_DEFAULT,
+                                            &d3dtex_right,
+                                            NULL);
+        if (FAILED(hr)) {
+            msg_Err(vd, "Failed to create secondary texture. (hr=0x%lx)", hr);
+            IDirect3DSurface9_Release(d3dleft_surface);
+            IDirect3DTexture9_Release(d3dtex);
+            IDirect3DVertexBuffer9_Release(d3dvtc);
+            return VLC_EGENERIC;
+        }
+
+        /* nvidia stereo surface */
+        LPDIRECT3DSURFACE9 d3dnv_surface;
+        hr = IDirect3DDevice9_CreateRenderTarget(d3ddev,
+                                                 sys->rect_display.right * 2,
+                                                 sys->rect_display.bottom + 1,
+                                                 D3DFMT_A8R8G8B8,
+                                                 D3DMULTISAMPLE_NONE,
+                                                 0,
+                                                 TRUE,
+                                                 &d3dnv_surface,
+                                                 NULL);
+        if (FAILED(hr)) {
+            msg_Err(vd, "Failed to create stereo surface. (hr=0x%lx)", hr);
+            IDirect3DTexture9_Release(d3dtex_right);
+            IDirect3DSurface9_Release(d3dleft_surface);
+            IDirect3DTexture9_Release(d3dtex);
+            IDirect3DVertexBuffer9_Release(d3dvtc);
+            return VLC_EGENERIC;
+        }                                       
+
+        LPDIRECT3DSURFACE9 d3dright_surface;
+        hr = IDirect3DTexture9_GetSurfaceLevel(d3dtex_right, 0, &d3dright_surface);
+        if (FAILED(hr)) {
+            msg_Err(vd, "Failed to get surface of right texture. (hr=0x%lx)", hr);
+            IDirect3DSurface9_Release(d3dnv_surface);
+            IDirect3DTexture9_Release(d3dtex_right);
+            IDirect3DSurface9_Release(d3dleft_surface);
+            IDirect3DTexture9_Release(d3dtex);
+            IDirect3DVertexBuffer9_Release(d3dvtc);
+            return VLC_EGENERIC;
+        }                                             
+       
+
+        LPDIRECT3DSURFACE9 d3dback_buffer;
+        hr = IDirect3DDevice9_GetBackBuffer(d3ddev,
+                                            0,
+                                            0,
+                                            D3DBACKBUFFER_TYPE_MONO,
+                                            &d3dback_buffer);
+        if (FAILED(hr)) {
+            msg_Err(vd, "Failed to get device back buffer. (hr=0x%lx)", hr);
+            IDirect3DSurface9_Release(d3dright_surface);
+            IDirect3DSurface9_Release(d3dleft_surface);
+            IDirect3DSurface9_Release(d3dnv_surface);
+            IDirect3DTexture9_Release(d3dtex_right);
+            IDirect3DTexture9_Release(d3dtex);
+            IDirect3DVertexBuffer9_Release(d3dvtc);
+            return VLC_EGENERIC;
+        }                        
+
+        sys->d3dtex_right = d3dtex_right;
+        sys->d3dnv_surface = d3dnv_surface;
+        sys->d3dback_buffer = d3dback_buffer;
+        sys->d3dright_surface = d3dright_surface;
+
+        sys->nvRectLeft.left = 0;
+        sys->nvRectLeft.top = 0;
+        sys->nvRectLeft.right = fmt->i_width;
+        sys->nvRectLeft.bottom = fmt->i_height;
+
+        sys->nvRectRight.left = fmt->i_width;
+        sys->nvRectRight.top = 0;
+        sys->nvRectRight.right = fmt->i_width * 2;
+        sys->nvRectRight.bottom = fmt->i_height;
+
+        sys->nvRectHeader.left = 0;
+        sys->nvRectHeader.top = 0;
+        sys->nvRectHeader.right = fmt->i_width * 2;
+        sys->nvRectHeader.bottom = fmt->i_height + 1;
+    }
+
+
     /* */
     sys->d3dtex = d3dtex;
     sys->d3dvtc = d3dvtc;
-
     sys->d3dregion_count = 0;
     sys->d3dregion       = NULL;
+    sys->d3dleft_surface = d3dleft_surface;
+    sys->left_tex_filled = false;
+    sys->right_tex_filled = false;
 
     sys->clear_scene = true;
 
@@ -1094,6 +1407,37 @@ static void Direct3DDestroyScene(vout_display_t *vd)
 
     Direct3DDeleteRegions(sys->d3dregion_count, sys->d3dregion);
 
+
+    if (sys->use_nvidia_3d)
+    {
+        LPDIRECT3DSURFACE9 d3dright_surface = sys->d3dright_surface;
+        if(d3dright_surface)
+            IDirect3DSurface9_Release(d3dright_surface);
+        sys->d3dright_surface = NULL;
+
+        LPDIRECT3DTEXTURE9 d3dtex_right = sys->d3dtex_right;
+        if(d3dtex_right)
+            IDirect3DTexture9_Release(d3dtex_right);
+        sys->d3dtex_right = NULL;
+
+        LPDIRECT3DSURFACE9 d3dnv_surface = sys->d3dnv_surface;
+        if(d3dnv_surface)
+            IDirect3DSurface9_Release(d3dnv_surface);
+        sys->d3dnv_surface = NULL;
+
+        LPDIRECT3DSURFACE9 d3dback_buffer = sys->d3dback_buffer;
+        if(d3dback_buffer)
+            IDirect3DSurface9_Release(d3dback_buffer);
+        sys->d3dback_buffer = NULL;
+    }
+
+
+    LPDIRECT3DSURFACE9 d3dleft_surface = sys->d3dleft_surface;
+    if(d3dleft_surface)
+        IDirect3DSurface9_Release(d3dleft_surface);
+    sys->d3dleft_surface = NULL;
+
+
     LPDIRECT3DVERTEXBUFFER9 d3dvtc = sys->d3dvtc;
     if (d3dvtc)
         IDirect3DVertexBuffer9_Release(d3dvtc);
@@ -1156,7 +1500,8 @@ static void Direct3DSetupVertices(CUSTOMVERTEX *vertices,
  */
 static int Direct3DImportPicture(vout_display_t *vd,
                                  d3d_region_t *region,
-                                 LPDIRECT3DSURFACE9 source)
+                                 LPDIRECT3DSURFACE9 source,
+                                 int i_eye)
 {
     vout_display_sys_t *sys = vd->sys;
     HRESULT hr;
@@ -1166,24 +1511,46 @@ static int Direct3DImportPicture(vout_display_t *vd,
         return VLC_EGENERIC;
     }
 
+
+    /* msg_Dbg(vd, "Receiving eye (i_eye=%i)", i_eye); */
+
     /* retrieve texture top-level surface */
     LPDIRECT3DSURFACE9 destination;
-    hr = IDirect3DTexture9_GetSurfaceLevel(sys->d3dtex, 0, &destination);
-    if (FAILED(hr)) {
-        msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
-        return VLC_EGENERIC;
+    /* either 2d, left, or no 3d */
+    if(!sys->use_nvidia_3d || i_eye < 2 )
+    {
+        destination = sys->d3dleft_surface;
+        if(i_eye == 0)
+        {
+            /* clear both eyes if we just received a 2d image
+               so we go back to drawing in 2d */
+            sys->right_tex_filled = false;
+            sys->left_tex_filled = false;
+        }
+        else
+            sys->left_tex_filled = true;
+    }
+    else
+    {
+        destination = sys->d3dright_surface;
+        sys->right_tex_filled = true;
     }
 
     /* Copy picture surface into texture surface
      * color space conversion happen here */
     hr = IDirect3DDevice9_StretchRect(sys->d3ddev, source, NULL, destination, NULL, D3DTEXF_NONE);
-    IDirect3DSurface9_Release(destination);
+
     if (FAILED(hr)) {
         msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
         return VLC_EGENERIC;
     }
 
     /* */
+/*    if(sys->left_tex_filled && sys->right_tex_filled &&
+        sys->use_nvidia_3d)
+        return VLC_SUCCESS;*/
+
+    /* only needed for 3d, but if we rapidly turn off 2d we require this to exist */
     region->texture = sys->d3dtex;
     Direct3DSetupVertices(region->vertex,
                           vd->sys->rect_src,
@@ -1304,55 +1671,121 @@ static int Direct3DRenderRegion(vout_display_t *vd,
     vout_display_sys_t *sys = vd->sys;
 
     LPDIRECT3DDEVICE9 d3ddev = vd->sys->d3ddev;
+    bool drawInStereo;
+
+    /* check if we can draw in stereo */    
+    if(sys->left_tex_filled && sys->right_tex_filled &&
+        sys->use_nvidia_3d)
+    {
+        /* check to see if stereo is turned on */
+        NvAPI_Status (WINAPI *NvAPIWrapper_Stereo_IsActivated)(
+	StereoHandle stereoHandle, NvU8 *pIsStereoOn);
+        NvAPIWrapper_Stereo_IsActivated = sys->nvStereoIsActivatedFunc;
+
+        NvU8 stereoOn;
+        if(NvAPIWrapper_Stereo_IsActivated(sys->nvStereoHandle, &stereoOn) == NVAPI_OK)
+            drawInStereo = (bool)stereoOn;
+        else
+            drawInStereo = false;
+    }
+    else
+        drawInStereo = false;
+
+    if(drawInStereo)
+    {
+        /* draw stereoscopic image */
+
+        RECT dest = sys->rect_dest_clipped;
+        /* copy left and right image to stereo surface*/
+        IDirect3DDevice9_StretchRect(sys->d3ddev, sys->d3dleft_surface,
+            &sys->nvRectLeft, sys->d3dnv_surface, &dest,
+            D3DTEXF_NONE);
+
+        dest.left += sys->rect_display.right;
+        dest.right += sys->rect_display.right;
+        IDirect3DDevice9_StretchRect(sys->d3ddev, sys->d3dright_surface,
+            &sys->nvRectLeft, sys->d3dnv_surface, &dest,
+            D3DTEXF_NONE);
+
+        /* write in the nvidia header */
+        D3DLOCKED_RECT lr;
+        IDirect3DSurface9_LockRect(sys->d3dnv_surface, &lr, NULL, 0);
+        LPNVSTEREOIMAGEHEADER pSIH =
+            (LPNVSTEREOIMAGEHEADER)(((unsigned char *) lr.pBits) +
+            (lr.Pitch * sys->rect_display.bottom));
+
+        /* write signature into stereo surface */
+        pSIH->dwSignature = NVSTEREO_IMAGE_SIGNATURE;
+        pSIH->dwBPP = 32;
+        pSIH->dwFlags = DXNV_SWAP_EYES;
+        pSIH->dwWidth = sys->rect_display.right * 2; 
+        pSIH->dwHeight = sys->rect_display.bottom;
+
+        IDirect3DSurface9_UnlockRect(sys->d3dnv_surface);
+
+        /* copy stereo surface into back buffer */
+        RECT stereo_source;
+        stereo_source.top = 0;
+        stereo_source.left = 0;
+        stereo_source.right = sys->rect_display.right * 2;
+        stereo_source.bottom = sys->rect_display.bottom + 1;
+        IDirect3DDevice9_StretchRect(sys->d3ddev, sys->d3dnv_surface,
+            &stereo_source, sys->d3dback_buffer,
+            &sys->rect_display, D3DTEXF_NONE);
+    }
+    else
+    {
+        /* draw 2d image */
 
-    LPDIRECT3DVERTEXBUFFER9 d3dvtc = sys->d3dvtc;
-    LPDIRECT3DTEXTURE9      d3dtex = region->texture;
+        LPDIRECT3DVERTEXBUFFER9 d3dvtc = sys->d3dvtc;
+        LPDIRECT3DTEXTURE9      d3dtex = region->texture;
 
-    HRESULT hr;
+        HRESULT hr;
 
-    /* Import vertices */
-    void *vertex;
-    hr = IDirect3DVertexBuffer9_Lock(d3dvtc, 0, 0, &vertex, D3DLOCK_DISCARD);
-    if (FAILED(hr)) {
-        msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
-        return -1;
-    }
-    memcpy(vertex, region->vertex, sizeof(region->vertex));
-    hr = IDirect3DVertexBuffer9_Unlock(d3dvtc);
-    if (FAILED(hr)) {
-        msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
-        return -1;
-    }
+        /* Import vertices */
+        void *vertex;
+        hr = IDirect3DVertexBuffer9_Lock(d3dvtc, 0, 0, &vertex, D3DLOCK_DISCARD);
+        if (FAILED(hr)) {
+            msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
+            return -1;
+        }
+        memcpy(vertex, region->vertex, sizeof(region->vertex));
+        hr = IDirect3DVertexBuffer9_Unlock(d3dvtc);
+        if (FAILED(hr)) {
+            msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
+            return -1;
+        }
 
-    // Setup our texture. Using textures introduces the texture stage states,
-    // which govern how textures get blended together (in the case of multiple
-    // textures) and lighting information. In this case, we are modulating
-    // (blending) our texture with the diffuse color of the vertices.
-    hr = IDirect3DDevice9_SetTexture(d3ddev, 0, (LPDIRECT3DBASETEXTURE9)d3dtex);
-    if (FAILED(hr)) {
-        msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
-        return -1;
-    }
+        // Setup our texture. Using textures introduces the texture stage states,
+        // which govern how textures get blended together (in the case of multiple
+        // textures) and lighting information. In this case, we are modulating
+        // (blending) our texture with the diffuse color of the vertices.
+        hr = IDirect3DDevice9_SetTexture(d3ddev, 0, (LPDIRECT3DBASETEXTURE9)d3dtex);
+        if (FAILED(hr)) {
+            msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
+            return -1;
+        }
 
-    // Render the vertex buffer contents
-    hr = IDirect3DDevice9_SetStreamSource(d3ddev, 0, d3dvtc, 0, sizeof(CUSTOMVERTEX));
-    if (FAILED(hr)) {
-        msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
-        return -1;
-    }
+        // Render the vertex buffer contents
+        hr = IDirect3DDevice9_SetStreamSource(d3ddev, 0, d3dvtc, 0, sizeof(CUSTOMVERTEX));
+        if (FAILED(hr)) {
+            msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
+            return -1;
+        }
 
-    // we use FVF instead of vertex shader
-    hr = IDirect3DDevice9_SetFVF(d3ddev, D3DFVF_CUSTOMVERTEX);
-    if (FAILED(hr)) {
-        msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
-        return -1;
-    }
+        // we use FVF instead of vertex shader
+        hr = IDirect3DDevice9_SetFVF(d3ddev, D3DFVF_CUSTOMVERTEX);
+        if (FAILED(hr)) {
+            msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
+            return -1;
+        }
 
-    // draw rectangle
-    hr = IDirect3DDevice9_DrawPrimitive(d3ddev, D3DPT_TRIANGLEFAN, 0, 2);
-    if (FAILED(hr)) {
-        msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
-        return -1;
+        // draw rectangle
+        hr = IDirect3DDevice9_DrawPrimitive(d3ddev, D3DPT_TRIANGLEFAN, 0, 2);
+        if (FAILED(hr)) {
+            msg_Dbg(vd, "%s:%d (hr=0x%0lX)", __FUNCTION__, __LINE__, hr);
+            return -1;
+        }
     }
     return 0;
 }
-- 
1.7.4.msysgit.0

